#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
ib_fundamental utility functions
"""

import dataclasses
import datetime
import json
import re
from typing import Any, Optional

from ib_async import FundamentalRatios
from pandas import DataFrame

re_pattern = re.compile(r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])")


def to_dataframe(table: Any, key: Optional[str] = None) -> DataFrame:
    """converts a list of dicts to a data frame
    Args:
        table (list(dict())): list of dicts/dataclasses
        key (str, optional): index dict key.
    Returns:
        pd.DataFrame: dataframe
    """
    df = DataFrame(table)
    if key is not None:
        df = df.set_index(keys=key, drop=True).sort_index()
    return df


def camel_to_snake(camel: str) -> str:
    """Convert CamelCase to snake_case"""
    # https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    return re_pattern.sub("_", camel).lower()


def to_json(obj: Any, **kwargs: Any) -> str:
    """Convert FundamentalData attributes to JSON"""

    class EnhancedJSONEncoder(json.JSONEncoder):
        """JSON encoder for dataclasses and datetime"""

        def default(self, o):
            if dataclasses.is_dataclass(o):
                return dataclasses.asdict(o)
            elif isinstance(o, (datetime.date, datetime.datetime)):
                return o.isoformat()
            elif isinstance(o, FundamentalRatios):
                return vars(o)

            return super().default(o)

    return json.dumps(obj, cls=EnhancedJSONEncoder, **kwargs)



#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Company Fundamental class
"""
# pylint: disable=attribute-defined-outside-init
# pylint: disable=missing-function-docstring

__all__ = [
    "FundamentalData",
]

from datetime import datetime
from typing import Optional

import pandas as pd
from ib_async import IB, Dividends, FundamentalRatios, Stock, Ticker
from pandas import DataFrame

from ib_fundamental.objects import (
    AnalystForecast,
    BalanceSheetSet,
    CashFlowSet,
    CompanyInfo,
    Dividend,
    DividendPerShare,
    EarningsPerShare,
    ForwardYear,
    IncomeSet,
    OwnershipReport,
    RatioSnapshot,
    Revenue,
    StatementCode,
    StatementData,
    statement_type,
)
from ib_fundamental.utils import to_dataframe

from .ib_client import IBClient
from .xml_parser import XMLParser

fromisoformat = datetime.fromisoformat


class FundamentalData:
    """Company fundamental data"""

    # pylint: disable=too-many-arguments,too-many-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, symbol: str, ib: IB) -> None:
        """Args:
        symbol (str): Company symbol/ticker
        ib (Optional[IB], optional): IB instance. Defaults to None.
        """
        self.client = IBClient(symbol=symbol, ib=ib)
        self.symbol = symbol
        self.contract: Stock = self.client.contract
        self.ticker: Optional[Ticker] = None
        self.parser = XMLParser(ib_client=self.client)

    def __repr__(self):
        cls_name = self.__class__.__qualname__
        return f"{cls_name}(symbol={self.symbol!r},IB={self.client.ib!r})"

    def __enter__(self):
        return self

    @property
    def income_annual(self) -> IncomeSet:
        """
        income_annual
        """
        try:
            return self.__income_annual
        except AttributeError:
            self.__income_annual = self.parser.get_fin_statement(
                statement="INC", period="annual"
            )
            return self.__income_annual

    @property
    def income_quarter(self) -> IncomeSet:
        """income_quarter"""
        try:
            return self.__income_quarter
        except AttributeError:
            self.__income_quarter = self.parser.get_fin_statement(
                statement="INC", period="quarter"
            )
            return self.__income_quarter

    @property
    def balance_annual(self) -> BalanceSheetSet:
        try:
            return self.__balance_annual
        except AttributeError:
            self.__balance_annual = self.parser.get_fin_statement(
                statement="BAL", period="annual"
            )
            return self.__balance_annual

    @property
    def balance_quarter(self) -> BalanceSheetSet:
        try:
            return self.__balance_quarter
        except AttributeError:
            self.__balance_quarter = self.parser.get_fin_statement(
                statement="BAL", period="quarter"
            )
            return self.__balance_quarter

    @property
    def cashflow_annual(self) -> CashFlowSet:
        try:
            return self.__cashflow_annual
        except AttributeError:
            self.__cashflow_annual = self.parser.get_fin_statement(
                statement="CAS", period="annual"
            )
            return self.__cashflow_annual

    @property
    def cashflow_quarter(self) -> CashFlowSet:
        try:
            return self.__cashflow_quarter
        except AttributeError:
            self.__cashflow_quarter = self.parser.get_fin_statement(
                statement="CAS", period="quarter"
            )
            return self.__cashflow_quarter

    @property
    def ownership_report(self) -> OwnershipReport:
        """Ownership Report"""
        try:
            return self.__ownership_report
        except AttributeError:
            self.__ownership_report = self.parser.get_ownership_report()
            return self.__ownership_report

    @property
    def dividend(self) -> list[Dividend] | None:
        try:
            return self.__dividend
        except AttributeError:
            self.__dividend: list[Dividend] | None = self.parser.get_dividend()
            return self.__dividend

    @property
    def div_ps_q(self) -> list[DividendPerShare] | None:
        try:
            return self.__div_ps_q
        except AttributeError:
            self.__div_ps_q: list[DividendPerShare] | None = (
                self.parser.get_div_per_share(report_type="R", period="3M")
            )
            return self.__div_ps_q

    @property
    def div_ps_ttm(self) -> list[DividendPerShare] | None:
        try:
            return self.__div_ps_ttm
        except AttributeError:
            self.__div_ps_ttm: list[DividendPerShare] | None = (
                self.parser.get_div_per_share(report_type="TTM")
            )
            return self.__div_ps_ttm

    @property
    def revenue_ttm(self) -> list[Revenue]:
        try:
            return self.__revenue_ttm
        except AttributeError:
            self.__revenue_ttm = self.parser.get_revenue(report_type="TTM")
            return self.__revenue_ttm

    @property
    def revenue_q(self) -> list[Revenue]:
        try:
            return self.__revenue_q
        except AttributeError:
            self.__revenue_q = self.parser.get_revenue(report_type="R", period="3M")
            return self.__revenue_q

    @property
    def eps_ttm(self) -> list[EarningsPerShare]:
        try:
            return self.__eps_ttm
        except AttributeError:
            self.__eps_ttm = self.parser.get_eps(report_type="TTM")
            return self.__eps_ttm

    @property
    def eps_q(self) -> list[EarningsPerShare]:
        try:
            return self.__eps_q
        except AttributeError:
            self.__eps_q = self.parser.get_eps(report_type="R", period="3M")
            return self.__eps_q

    @property
    def analyst_forecast(self) -> AnalystForecast:
        try:
            return self.__analyst_forecast
        except AttributeError:
            self.__analyst_forecast = self.parser.get_analyst_forecast()
            return self.__analyst_forecast

    @property
    def ratios(self) -> RatioSnapshot:
        try:
            return self.__ratios
        except AttributeError:
            self.__ratios = self.parser.get_ratios()
            return self.__ratios

    @property
    def fundamental_ratios(self) -> FundamentalRatios:
        try:
            return self.__fundamental_ratios
        except AttributeError:
            self.__fundamental_ratios = self.client.get_ratios()
            self.ticker = self.client.ib.ticker(self.contract)
            return self.__fundamental_ratios

    @property
    def dividend_summary(self) -> Dividends:
        try:
            return self.__dividend_summary
        except AttributeError:
            self.__dividend_summary = self.client.get_dividends()
            self.ticker = self.client.ib.ticker(self.contract)
            return self.__dividend_summary

    @property
    def fy_estimates(self) -> list[ForwardYear]:
        try:
            return self.__fy_estimates
        except AttributeError:
            self.__fy_estimates = self.parser.get_fy_estimates()
            return self.__fy_estimates

    @property
    def fy_actuals(self) -> list[ForwardYear]:
        try:
            return self.__fy_actuals
        except AttributeError:
            self.__fy_actuals = self.parser.get_fy_actuals()
            return self.__fy_actuals

    @property
    def company_info(self) -> CompanyInfo:
        try:
            return self.__company_info
        except AttributeError:
            self.__company_info = self.parser.get_company_info()
            return self.__company_info


class CompanyFinancials:
    """Company Financials"""

    def __init__(self, symbol: str, ib: IB) -> None:
        """_summary_

        Args:
            symbol (str): company symbol/ticker
            ib (ib_async.IB): ib_async.IB instance
        """
        self.data = FundamentalData(symbol=symbol, ib=ib)

    def __repr__(self):
        cls_name = self.__class__.__qualname__
        return f"{cls_name}(symbol={self.data.symbol!r},IB={self.data.client.ib!r})"

    def _get_data_frame(
        self,
        statement: StatementData,
    ) -> DataFrame:
        """Build dataframe for pp"""
        _df = to_dataframe(statement)
        return _df.T.sort_index(axis=1, ascending=False)  # sort columns

    def _get_map_items(self, stat_code: StatementCode) -> DataFrame:
        """build map items for pp"""
        _df = to_dataframe(self.data.parser.get_map_items(statement=stat_code))
        _df.coa_item = _df.coa_item.str.lower()
        return _df

    def _get_header(
        self, data: DataFrame, statement_code: StatementCode, idx: int = 6
    ) -> DataFrame:
        """build header for pp"""
        _header = data.iloc[:idx]
        _header = (
            _header.assign(line_id=range(idx))
            .assign(statement_type=statement_code)
            .reset_index()
            .rename(columns={"index": "map_item"})
        )
        return _header.assign(coa_item=_header["map_item"])

    def _join(
        self, data: DataFrame, header: DataFrame, mapping: DataFrame, idx: int
    ) -> DataFrame:
        """join data to present"""
        _pp = mapping.join(data, on="coa_item")
        _df = pd.concat([header, _pp]).set_index("line_id")

        (_names,) = _df.loc[
            _df["coa_item"] == "end_date", _df.columns[1:idx]
        ].values.tolist()
        _l = _df.columns.to_list()
        _l[1:idx] = _names
        _df.columns = _l
        _df.statement_type = _df.statement_type.map(lambda x: statement_type[x])
        _df = _df.drop(columns="coa_item").dropna()
        return _df

    def _build_statement(
        self, data: StatementData, statement_code: StatementCode, idx: int
    ) -> DataFrame:
        """build statement pp"""
        _map = self._get_map_items(stat_code=statement_code)
        _data = self._get_data_frame(statement=data)
        _header = self._get_header(data=_data, statement_code=statement_code)
        # pp
        return self._join(data=_data, header=_header, mapping=_map, idx=idx)

    @property
    def balance_quarter(self) -> DataFrame | None:
        """Quarterly balance statement"""
        if self.data.balance_quarter:
            return self._build_statement(self.data.balance_quarter, "BAL", 6)
        return None

    @property
    def balance_annual(self) -> DataFrame | None:
        if self.data.balance_annual:
            return self._build_statement(self.data.balance_annual, "BAL", 7)
        return None

    @property
    def income_quarter(self) -> DataFrame | None:
        if self.data.income_quarter:
            return self._build_statement(self.data.income_quarter, "INC", 6)
        return None

    @property
    def income_annual(self) -> DataFrame | None:
        if self.data.income_annual:
            return self._build_statement(self.data.income_annual, "INC", 7)
        return None

    @property
    def cashflow_quarter(self) -> DataFrame | None:
        if self.data.cashflow_annual:
            return self._build_statement(self.data.cashflow_quarter, "CAS", 6)
        return None

    @property
    def cashflow_annual(self) -> DataFrame | None:
        if self.data.cashflow_annual:
            return self._build_statement(self.data.cashflow_annual, "CAS", 7)
        return None

    @property
    def dividends(self) -> DataFrame | None:
        if self.data.dividend:
            return to_dataframe(self.data.dividend, key="ex_date")
        return None

    @property
    def dividends_ps_q(self) -> DataFrame | None:
        if self.data.div_ps_q:
            return to_dataframe(self.data.div_ps_q, key="as_of_date")
        return None

    @property
    def dividends_ps_ttm(self) -> DataFrame | None:
        if self.data.div_ps_ttm:
            return to_dataframe(self.data.div_ps_ttm, key="as_of_date")
        return None

    @property
    def revenue_q(self) -> DataFrame | None:
        if self.data.revenue_q:
            return to_dataframe(self.data.revenue_q, key="as_of_date")
        return None

    @property
    def revenue_ttm(self) -> DataFrame | None:
        if self.data.revenue_ttm:
            return to_dataframe(self.data.revenue_ttm, key="as_of_date")
        return None

    @property
    def eps_q(self) -> DataFrame | None:
        if self.data.eps_q:
            return to_dataframe(self.data.eps_q, key="as_of_date")
        return None

    @property
    def eps_ttm(self) -> DataFrame | None:
        if self.data.eps_ttm:
            return to_dataframe(self.data.eps_ttm, key="as_of_date")
        return None

    @property
    def ownership(self) -> DataFrame | None:
        if self.data.ownership_report:
            return to_dataframe(self.data.ownership_report.ownership_details)
        return None

    @property
    def fy_actuals(self) -> DataFrame | None:
        if self.data.fy_actuals:
            return to_dataframe(self.data.fy_actuals, key="updated")
        return None

    @property
    def fy_estimates(self) -> DataFrame | None:
        if self.data.fy_estimates:
            return to_dataframe(self.data.fy_estimates)
        return None

    @property
    def analyst_forecast(self) -> DataFrame | None:
        if self.data.analyst_forecast:
            return to_dataframe([self.data.analyst_forecast]).T
        return None

    @property
    def company_information(self) -> DataFrame | None:
        if self.data.company_info:
            return to_dataframe([self.data.company_info]).T
        return None

    @property
    def ratios(self) -> DataFrame | None:
        if self.data.ratios:
            return to_dataframe([self.data.ratios]).T
        return None

    @property
    def fundamental_ratios(self) -> DataFrame | None:
        if self.data.fundamental_ratios:
            return to_dataframe([vars(self.data.fundamental_ratios)]).T
        return None



#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Created on Thu May 9 16:21:58 2021

@author: gonzo
"""
# pylint: disable=too-many-instance-attributes
from dataclasses import dataclass
from datetime import date, datetime
from typing import Literal, Optional, Union

StatementCode = Literal["INC", "BAL", "CAS"]
PeriodType = Literal["annual", "quarter"]
StatementPeriod = Literal["Interim", "Annual"]
ReportType = Literal[
    "ReportsFinStatements",
    "ReportsFinSummary",
    "ReportSnapshot",
    "RESC",
    "ReportsOwnership",
    "CalendarReport",
]

statement_type = {
    "INC": "Income Statement",
    "BAL": "Balance Sheet Statement",
    "CAS": "Cash Flow Statement",
}


@dataclass
class FinancialStatement:
    """Financial Statement Base class"""

    period: Optional[StatementPeriod] = None
    end_date: Optional[date] = None
    fiscal_year: Optional[int] = None
    period_number: Optional[int] = None
    date_10Q: str = ""  # pylint: disable=invalid-name
    date_10K: str = ""  # pylint: disable=invalid-name


@dataclass
class IncomeStatement(FinancialStatement):
    """Income Statement"""

    statement = "Income Statement"
    srev: Optional[float] = None
    spre: Optional[float] = None
    rnii: Optional[float] = None
    sore: Optional[float] = None
    rtlr: Optional[float] = None
    slba: Optional[float] = None
    epac: Optional[float] = None
    rrgl: Optional[float] = None
    scor: Optional[float] = None
    sgrp: Optional[float] = None
    ssga: Optional[float] = None
    siib: Optional[float] = None
    stie: Optional[float] = None
    enii: Optional[float] = None
    erad: Optional[float] = None
    sdpr: Optional[float] = None
    sinn: Optional[float] = None
    suie: Optional[float] = None
    sooe: Optional[float] = None
    etoe: Optional[float] = None
    sopi: Optional[float] = None
    snin: Optional[float] = None
    ngla: Optional[float] = None
    sont: Optional[float] = None
    eibt: Optional[float] = None
    ttax: Optional[float] = None
    tiat: Optional[float] = None
    cmin: Optional[float] = None
    ceia: Optional[float] = None
    cgap: Optional[float] = None
    nibx: Optional[float] = None
    stxi: Optional[float] = None
    ninc: Optional[float] = None
    sani: Optional[float] = None
    ciac: Optional[float] = None
    xnic: Optional[float] = None
    sdaj: Optional[float] = None
    sdni: Optional[float] = None
    sdws: Optional[float] = None
    sdbf: Optional[float] = None
    ddps1: Optional[float] = None
    vdes: Optional[float] = None
    ellp: Optional[float] = None
    siap: Optional[float] = None
    snii: Optional[float] = None
    snie: Optional[float] = None


@dataclass
class BalanceSheetStatement(FinancialStatement):
    """Balance Sheet class"""

    statement = "Balance Sheet Statement"
    acsh: Optional[float] = None
    acae: Optional[float] = None
    asti: Optional[float] = None
    acdb: Optional[float] = None
    scsi: Optional[float] = None
    aacr: Optional[float] = None
    atrc: Optional[float] = None
    aitl: Optional[float] = None
    soea: Optional[float] = None
    antl: Optional[float] = None
    appy: Optional[float] = None
    soca: Optional[float] = None
    atca: Optional[float] = None
    aptc: Optional[float] = None
    adep: Optional[float] = None
    appn: Optional[float] = None
    agwi: Optional[float] = None
    aint: Optional[float] = None
    sinv: Optional[float] = None
    apre: Optional[float] = None
    adpa: Optional[float] = None
    altr: Optional[float] = None
    sola: Optional[float] = None
    soat: Optional[float] = None
    spol: Optional[float] = None
    atot: Optional[float] = None
    lapb: Optional[float] = None
    lpba: Optional[float] = None
    laex: Optional[float] = None
    lstd: Optional[float] = None
    lcld: Optional[float] = None
    ldbt: Optional[float] = None
    sobl: Optional[float] = None
    socl: Optional[float] = None
    ltcl: Optional[float] = None
    lltd: Optional[float] = None
    lclo: Optional[float] = None
    lttd: Optional[float] = None
    stld: Optional[float] = None
    sbdt: Optional[float] = None
    lmin: Optional[float] = None
    sltl: Optional[float] = None
    ltll: Optional[float] = None
    srpr: Optional[float] = None
    sprs: Optional[float] = None
    scms: Optional[float] = None
    qpic: Optional[float] = None
    qred: Optional[float] = None
    qtsc: Optional[float] = None
    qedg: Optional[float] = None
    qugl: Optional[float] = None
    sote: Optional[float] = None
    qtle: Optional[float] = None
    qtel: Optional[float] = None
    qtco: Optional[float] = None
    qtpo: Optional[float] = None
    stbp: Optional[float] = None
    lstb: Optional[float] = None


@dataclass
class CashFlowStatement(FinancialStatement):
    """Cash flow statement class"""

    statement = "Cash Flow Statement"
    onet: Optional[float] = None
    sded: Optional[float] = None
    samt: Optional[float] = None
    obdt: Optional[float] = None
    snci: Optional[float] = None
    socf: Optional[float] = None
    otlo: Optional[float] = None
    scex: Optional[float] = None
    sicf: Optional[float] = None
    itli: Optional[float] = None
    sfcf: Optional[float] = None
    fcdp: Optional[float] = None
    fpss: Optional[float] = None
    fprd: Optional[float] = None
    ftlf: Optional[float] = None
    sfee: Optional[float] = None
    sncc: Optional[float] = None
    scip: Optional[float] = None
    sctp: Optional[float] = None


BalanceSheetSet = list[BalanceSheetStatement]
IncomeSet = list[IncomeStatement]
CashFlowSet = list[CashFlowStatement]
StatementData = Union[BalanceSheetSet, IncomeSet, CashFlowSet]


@dataclass
class StatementMap:
    """Financial statement map item"""

    coa_item: str
    map_item: str
    statement_type: StatementCode
    line_id: int


StatementMapping = list[StatementMap]

statement_map = {
    "INC": IncomeStatement,
    "BAL": BalanceSheetStatement,
    "CAS": CashFlowStatement,
}


@dataclass
class Dividend:
    """Dividend"""

    type: str
    ex_date: datetime
    record_date: datetime
    pay_date: datetime
    declaration_date: datetime
    currency: str
    value: float


@dataclass
class DividendPerShare:
    """Dividend per share"""

    as_of_date: str
    report_type: str
    period: str
    currency: str
    value: float


@dataclass
class Revenue:
    """Revenue"""

    as_of_date: datetime
    report_type: str
    period: str
    currency: str
    revenue: float


@dataclass
class EarningsPerShare:
    """Earnings per share"""

    as_of_date: datetime
    report_type: str
    period: str
    currency: str
    eps: float


@dataclass
class AnalystForecast:
    """Analyst Forecast"""

    cons_recom: float
    target_price: float
    proj_lt_growth_rate: float
    proj_pe: float
    proj_sales: float
    proj_sales_q: float
    proj_eps: float
    proj_epsq: float
    proj_profit: float
    proj_dps: float


@dataclass
class RatioSnapshot:
    """Company Ratios"""

    nprice: float
    nhig: float
    nlow: float
    pdate: date
    vol10davg: float
    ev: float
    mktcap: float
    ttmrev: float
    ttmebitd: float
    ttmniac: float
    ttmepsxclx: float
    ttmrevps: float
    qbvps: float
    qcshps: float
    ttmcfshr: float
    ttmdivshr: float
    ttmgrosmgn: float
    ttmroepct: float
    ttmpr2rev: float
    peexclxor: float
    price2bk: float
    employees: float


@dataclass
class ForwardYear:
    """Forward Year"""

    type: Literal["Estimate", "Actual"]
    item: str  # type
    unit: Literal["U", "M"]
    period_type: Literal["A", "Q"]
    fyear: int
    end_month: int
    end_cal_year: int
    value: float
    est_type: Literal["High", "Low", "Mean", "Median", "NumOfEst", "StdDev", ""] = ""
    updated: Optional[datetime] = None


@dataclass
class CompanyInfo:
    """Company Info"""

    ticker: str
    company_name: str
    cik: str
    exchange_code: str
    exchange: str
    irs: str


@dataclass
class OwnershipCompany:
    """Ownership Company details"""

    ISIN: str  # pylint: disable=invalid-name
    float_shares: int
    as_of_date: datetime


@dataclass
class OwnershipDetails:
    """Ownership details"""

    owner_id: str
    type: int
    as_of_date: datetime
    name: str
    quantity: float
    currency: Optional[str] = ""


@dataclass
class OwnershipReport:
    """Ownership Report"""

    company: OwnershipCompany
    ownership_details: list[OwnershipDetails]



#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Created on Fri Apr 30 16:21:58 2021

@author: gonzo
"""
# pylint: disable=attribute-defined-outside-init
from xml.etree.ElementTree import Element

from defusedxml.ElementTree import fromstring

from .ib_client import IBClient

__all__ = [
    "XMLReport",
]


class XMLReport:
    """XML Report Cache"""

    def __init__(self, ib_client: IBClient):
        self.client = ib_client

    def __repr__(self):
        cls_name = self.__class__.__qualname__
        return f"{cls_name}(ib_client={self.client!r})"

    @property
    def fin_statements(self) -> Element:
        """Request financial statements"""
        try:
            return self.__fin_statements
        except AttributeError:
            self.__fin_statements: Element = fromstring(
                self.client.ib_req_fund("ReportsFinStatements")
            )
            return self.__fin_statements

    @property
    def fin_summary(self) -> Element:
        """request financial summary"""
        try:
            return self.__fin_summary
        except AttributeError:
            self.__fin_summary: Element = fromstring(
                self.client.ib_req_fund("ReportsFinSummary")
            )
            return self.__fin_summary

    @property
    def snapshot(self) -> Element:
        """request snapshot report"""
        try:
            return self.__snapshot
        except AttributeError:
            self.__snapshot: Element = fromstring(
                self.client.ib_req_fund("ReportSnapshot")
            )
            return self.__snapshot

    @property
    def resc(self) -> Element:
        """request RESC"""
        try:
            return self.__resc
        except AttributeError:
            self.__resc: Element = fromstring(self.client.ib_req_fund("RESC"))
            return self.__resc

    @property
    def ownership(self) -> Element:
        """request ReportsOwnership"""
        try:
            return self.__ownership
        except AttributeError:
            self.__ownership: Element = fromstring(
                self.client.ib_req_fund("ReportsOwnership")
            )
            return self.__ownership

    @property
    def calendar(self) -> Element:
        """Calendar Report"""
        try:
            return self.__calendar
        except AttributeError:
            self.__calendar: Element = fromstring(
                self.client.ib_req_fund("CalendarReport")
            )
            return self.__calendar



#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Created on Fri Apr 30 16:21:58 2021

@author: gonzo
"""
__all__ = [
    "XMLParser",
]

from datetime import date, datetime
from typing import Literal, Optional

import pandas as pd

from .ib_client import IBClient
from .objects import (
    AnalystForecast,
    CompanyInfo,
    Dividend,
    DividendPerShare,
    EarningsPerShare,
    ForwardYear,
    OwnershipCompany,
    OwnershipDetails,
    OwnershipReport,
    PeriodType,
    RatioSnapshot,
    Revenue,
    StatementCode,
    StatementMap,
    StatementMapping,
    statement_map,
)
from .utils import camel_to_snake
from .xml_report import XMLReport

fromisoformat = datetime.fromisoformat

SummaryReportType = Literal["A", "TTM", "R", "P", None]
SummaryPeriod = Literal["12M", "3M", None]


class XMLParser:
    """Parser for IBKR xml company fundamental data"""

    def __init__(self, ib_client: IBClient):
        self.xml_report = XMLReport(ib_client=ib_client)

    def __repr__(self):
        cls_name = self.__class__.__qualname__
        return f"{cls_name}(ib_client={self.xml_report.client!r}"

    def get_fin_statement(
        self,
        statement: StatementCode = "INC",
        period: PeriodType = "annual",
        end_date: Optional[date] = None,
    ):
        """

        Parameters
        ----------
        period : 'annual' OR 'quarter', mandatory.
            statement period, annual or quarter. The default is 'annual'.
        statement : 'INC', 'BAL'or 'CAS'. mandatory.
            the statement to be generated
        end_date : 'YYYY-MM-DD' str format, optional
            statement date. The default is None.

        Raises
        ------
        Exception
            if period parameter is missing

        Returns
        -------
        list
            DESCRIPTION.

        """
        xpath = {
            "annual": ".//AnnualPeriods/FiscalPeriod",
            "quarter": ".//InterimPeriods/FiscalPeriod",
            "xp_fiscal_period": '[@EndDate="{0}"]/Statement[@Type="{1}"]/lineItem',
        }

        # statement type
        xp_statement = statement
        # period type
        xp_line = xpath[period]
        # filter by end_date
        if end_date is not None:
            xp_line_ed = xp_line + f'[@EndDate="{end_date.isoformat()}"]'
        else:
            xp_line_ed = xp_line

        fp = self.xml_report.fin_statements.findall(xp_line_ed)

        if period == "quarter":
            fperiods = [
                {
                    "period": fperiod.attrib["Type"],
                    "end_date": fperiod.attrib["EndDate"],
                    "fiscal_year": fperiod.attrib["FiscalYear"],
                    "period_number": fperiod.attrib["FiscalPeriodNumber"],
                    "date_10Q": fperiod.find(".//Source").attrib["Date"],
                }
                for fperiod in fp
            ]
        else:
            fperiods = [
                {
                    "period": fperiod.attrib["Type"],
                    "end_date": fperiod.attrib["EndDate"],
                    "fiscal_year": fperiod.attrib["FiscalYear"],
                    "date_10K": fperiod.find(".//Source").attrib["Date"],
                }
                for fperiod in fp
            ]

        fs = []

        for p in fperiods:
            if end_date is not None:
                ed = end_date.isoformat()
            else:
                ed = p["end_date"]

            xp = xpath["xp_fiscal_period"].format(ed, xp_statement)
            fs_line = self.xml_report.fin_statements.findall(xp_line + xp)

            fs.append({i.attrib["coaCode"].lower(): float(i.text) for i in fs_line})

        return [statement_map[statement](**i, **j) for i, j in zip(fperiods, fs)]

    def get_map_items(
        self, statement: Optional[StatementCode] = None
    ) -> StatementMapping:
        """
        mapItems

        Returns
        -------
        list of dict with FinStatement name mapping
        """

        fa = ".//COAMap/"
        fs = self.xml_report.fin_statements.findall(fa)

        _map_items: StatementMapping = [
            StatementMap(
                coa_item=mi.attrib["coaItem"],
                map_item=mi.text,
                statement_type=mi.attrib["statementType"],
                line_id=int(mi.attrib["lineID"]),
            )
            for mi in fs
            if statement is None
            or (statement is not None and mi.attrib["statementType"] == statement)
        ]
        return _map_items

    def get_ownership_report(self) -> OwnershipReport:
        """Ownership Report"""
        fs = self.xml_report.ownership
        (isin,) = fs.findall("./ISIN")
        (fa,) = fs.findall("./floatShares")
        company = OwnershipCompany(
            ISIN=isin.text,
            float_shares=int(fa.text),
            as_of_date=fromisoformat(fa.attrib["asofDate"]),
        )
        _l = []
        fa = fs.findall("./Owner")

        for i in fa:
            d = {}
            d["owner_id"] = i.attrib["ownerId"]

            for j in i.findall("./"):
                if "as_of_date" not in d and j.attrib:
                    d["as_of_date"] = datetime.fromisoformat(j.attrib["asofDate"])
                d[j.tag] = float(j.text) if j.tag == "quantity" else j.text

            _l.append(OwnershipDetails(**d))

        return OwnershipReport(company=company, ownership_details=_l)

    def get_dividend(self) -> list[Dividend] | None:
        """get dividends"""
        fa = "./Dividends"
        fs = self.xml_report.fin_summary.find(fa)
        if fs is None:
            return None
        curr = fs.attrib["currency"]
        _dividend = [
            Dividend(
                type=i.attrib["type"],
                ex_date=fromisoformat(i.attrib["exDate"]),
                record_date=fromisoformat(i.attrib["recordDate"]),
                pay_date=fromisoformat(i.attrib["payDate"]),
                declaration_date=fromisoformat(i.attrib["declarationDate"]),
                currency=curr,
                value=float(i.text),
            )
            for i in fs
        ]
        return _dividend

    def get_div_per_share(
        self,
        report_type: SummaryReportType = None,
        period: SummaryPeriod = None,
    ) -> list[DividendPerShare] | None:
        """Dividend per share"""
        fa = "./DividendPerShares"
        fs = self.xml_report.fin_summary.find(fa)
        if fs is None:
            return None
        curr = fs.attrib["currency"]

        _div_ps = [
            DividendPerShare(
                as_of_date=fromisoformat(i.attrib["asofDate"]),
                report_type=i.attrib["reportType"],
                period=i.attrib["period"],
                currency=curr,
                value=float(i.text),
            )
            for i in fs
            if (
                report_type is None
                or (report_type is not None and i.attrib["reportType"] == report_type)
            )
            and (
                period is None or (period is not None and i.attrib["period"] == period)
            )
        ]
        return _div_ps

    def get_revenue(
        self,
        report_type: SummaryReportType = None,
        period: SummaryPeriod = None,
    ) -> list[Revenue]:
        """Revenue"""
        fa = "./TotalRevenues"
        fs = self.xml_report.fin_summary.find(fa)
        curr = fs.attrib["currency"]

        _revenue = [
            Revenue(
                as_of_date=fromisoformat(tr.attrib["asofDate"]),
                report_type=tr.attrib["reportType"],
                period=tr.attrib["period"],
                currency=curr,
                revenue=float(tr.text),
            )
            for tr in fs
            if (
                report_type is None
                or (report_type is not None and tr.attrib["reportType"] == report_type)
            )
            and (
                period is None or (period is not None and tr.attrib["period"] == period)
            )
        ]
        return _revenue

    def get_eps(
        self,
        report_type: SummaryReportType = None,
        period: SummaryPeriod = None,
    ) -> list[EarningsPerShare]:
        """Earnings per share"""
        fa = "./EPSs"
        fs = self.xml_report.fin_summary.find(fa)
        curr = fs.attrib["currency"]

        _eps = [
            EarningsPerShare(
                as_of_date=fromisoformat(tr.attrib["asofDate"]),
                report_type=tr.attrib["reportType"],
                period=tr.attrib["period"],
                currency=curr,
                eps=float(tr.text),
            )
            for tr in fs
            if (
                report_type is None
                or (report_type is not None and tr.attrib["reportType"] == report_type)
            )
            and (
                period is None or (period is not None and tr.attrib["period"] == period)
            )
        ]
        return _eps

    def get_analyst_forecast(self) -> AnalystForecast:
        """Analyst forecast"""
        fa = ".//ForecastData/Ratio"
        fs = self.xml_report.snapshot.findall(fa)

        _analyst_forecast = AnalystForecast(
            **{
                camel_to_snake(g.attrib["FieldName"]): float(
                    g.findall("./Value")[0].text
                )
                for g in fs
            }
        )
        return _analyst_forecast

    def get_ratios(self) -> RatioSnapshot:
        """Company ratios snapshot"""
        fa = ".//Ratios/Group/Ratio"
        fs = self.xml_report.snapshot.findall(fa)

        _ratios = RatioSnapshot(
            **{
                r.attrib["FieldName"].lower(): (
                    fromisoformat(r.text) if r.attrib["Type"] == "D" else float(r.text)
                )
                for r in fs
            }
        )
        return _ratios

    def get_fy_estimates(self) -> list[ForwardYear]:
        """Forward Year estimates"""
        fs = self.xml_report.resc
        _fy_estimates = [
            ForwardYear(
                type="Estimate",
                item=a.attrib["type"],
                unit=a.attrib["unit"],
                period_type=p.attrib["periodType"],
                fyear=int(p.attrib["fYear"]),
                end_month=int(p.attrib["endMonth"]),
                end_cal_year=int(p.attrib["endCalYear"]),
                value=float(e[0].text),
                est_type=e.attrib["type"],
            )
            for a in fs.iter("FYEstimate")
            for p in a.iter("FYPeriod")
            for e in p.iter("ConsEstimate")
        ]
        return _fy_estimates

    def get_fy_actuals(self) -> list[ForwardYear]:
        """Forward year actuals"""
        fs = self.xml_report.resc
        _fy_actuals = [
            ForwardYear(
                type="Actual",
                item=a.attrib["type"],
                unit=a.attrib["unit"],
                period_type=p.attrib["periodType"],
                fyear=int(p.attrib["fYear"]),
                end_month=int(p.attrib["endMonth"]),
                end_cal_year=int(p.attrib["endCalYear"]),
                value=float(p[0].text),
                updated=pd.to_datetime(p[0].attrib["updated"]),
            )
            for a in fs.iter("FYActual")
            for p in a.iter("FYPeriod")
        ]
        return _fy_actuals

    def get_company_info(self) -> CompanyInfo:
        """Company Info"""
        fs = self.xml_report.fin_statements
        coids = {r.attrib["Type"]: r.text for r in fs.findall("./CoIDs/CoID")}
        issue_id = {
            r.attrib["Type"]: r.text for r in fs.findall("./Issues/Issue/IssueID")
        }
        exchange = {r.tag: r.text for r in fs.findall("./Issues/Issue/Exchange")}
        exchange_code = {
            "code": r.attrib["Code"] for r in fs.findall("./Issues/Issue/Exchange")
        }
        _company_info = CompanyInfo(
            ticker=issue_id["Ticker"],
            company_name=coids["CompanyName"],
            cik=coids["CIKNo"],
            exchange_code=exchange_code["code"],
            exchange=exchange["Exchange"],
            irs=coids["IRSNo"],
        )
        return _company_info



#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Created on Fri Apr 30 16:21:58 2021

@author: gonzo
"""

__all__ = [
    "IBClient",
]

from ib_async import IB, Dividends, FundamentalRatios, Stock, Ticker

from .objects import ReportType


class IBClient:
    """IB client"""

    ticker: Ticker
    tick_list: str = "258,456"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        symbol: str,
        ib: IB,
    ):

        self.ib = ib
        if symbol:
            self.contract: Stock = self.make_contract(symbol)
            self.symbol: str = symbol
        else:
            raise ValueError("No symbol defined.")

        if self.ib.isConnected():
            self.host = self.ib.client.host
            self.port = self.ib.client.port
            self.client_id = self.ib.client.clientId
        else:
            raise ValueError("IB is not connected.")

    def __repr__(self):
        cls_name = self.__class__.__qualname__
        return (
            f"{cls_name}(symbol={self.symbol!r},host={self.host!r},"
            f"port={self.port!r},client_id={self.client_id!r},IB={self.ib!r},"
            f"contract={self.contract!r})"
        )

    def __del__(self):
        self.cancel_ticket()

    def __enter__(self):
        return self

    def __exit__(self, *_exc):
        self.cancel_ticket()

    def make_contract(
        self, symbol: str, exchange: str = "SMART", currency: str = "USD"
    ) -> Stock:
        """Stock contract factory method

        Args:
            symbol (str): stock symbol
            exchange (str, optional): exchange. Defaults to "SMART".
            currency (str, optional): currency. Defaults to "USD".

        Returns:
            Contract: Stock contract
        """
        _stock = Stock(symbol=symbol, exchange=exchange, currency=currency)
        self.ib.qualifyContracts(_stock)
        return _stock

    def ib_req_fund(self, report_type: ReportType) -> str:
        """
        reportType:
            - 'ReportsFinSummary': Financial summary
            - 'ReportsOwnership': Company's ownership
            - 'ReportSnapshot': Company's financial overview
            - 'ReportsFinStatements': Financial Statements
            - 'RESC': Analyst Estimates
            - 'CalendarReport': Company's calendar
        """
        xml = self.ib.reqFundamentalData(self.contract, report_type)
        if xml:
            return xml
        raise ValueError(
            f"No response for report {report_type}, contract: {self.contract}"
        )

    def get_ticker(self) -> Ticker:
        """get ticker data"""
        self.ticker = self.ib.reqMktData(
            contract=self.contract,
            genericTickList=self.tick_list,
            snapshot=False,
        )
        return self.ticker

    def get_ratios(self) -> FundamentalRatios:
        """request market data ticker with fundamental ratios"""
        self.get_ticker()
        if self.ticker.fundamentalRatios is None:
            while self.ticker.fundamentalRatios is None:
                self.ib.sleep(0.0)
        return self.ticker.fundamentalRatios

    def get_dividends(self) -> Dividends:
        """get dividend information from ticker"""
        self.get_ticker()
        if self.ticker.dividends is None:
            while self.ticker.dividends is None:
                self.ib.sleep(0.0)
        return self.ticker.dividends

    def cancel_ticket(self) -> None:
        """cancel ticket market data"""
        if not hasattr(self, "ticker"):
            return None
        if self.ticker in self.ib.tickers() and self.is_connected():
            self.ib.cancelMktData(self.contract)
        return None

    def is_connected(self) -> bool:
        """is connected to TWS api"""
        return self.ib.isConnected()

    def disconnect(self) -> None:
        """Disconnect from TWS API"""
        self.ib.disconnect()
